"""
Built-in block-level token classes.
"""

import re
from itertools import zip_longest
from threading import local

import mistletoe.block_tokenizer as tokenizer
from mistletoe import span_token
from mistletoe.core_tokens import (
    follows,
    shift_whitespace,
    whitespace,
    is_control_char,
    normalize_label,
)


"""
Tokens to be included in the parsing process, in the order specified.
"""
__all__ = [
    "BlockCode",
    "Heading",
    "Quote",
    "CodeFence",
    "ThematicBreak",
    "List",
    "Table",
    "Footnote",
    "Paragraph",
]


"""
Stores a reference to the current document token.

When parsing, footnote entries will be stored in the document by
accessing this pointer.
"""
# TODO make thread local
_root_node = None


def tokenize(lines, start_line=0):
    """
    A wrapper around block_tokenizer.tokenize. Pass in all block-level
    token constructors as arguments to block_tokenizer.tokenize.

    Doing so (instead of importing block_token module in block_tokenizer)
    avoids cyclic dependency issues, and allows for future injections of
    custom token classes.

    _token_types variable is at the bottom of this module.

    See also: block_tokenizer.tokenize, span_token.tokenize_inner.
    """
    return tokenizer.tokenize(lines, _token_types.value, start_line)


def add_token(token_cls, position=0):
    """
    Allows external manipulation of the parsing process.
    This function is usually called in BaseRenderer.__enter__.

    Arguments:
        token_cls (SpanToken): token to be included in the parsing process.
        position (int): the position for the token class to be inserted into.
    """
    _token_types.value.insert(position, token_cls)


def remove_token(token_cls):
    """
    Allows external manipulation of the parsing process.
    This function is usually called in BaseRenderer.__exit__.

    Arguments:
        token_cls (BlockToken): token to be removed from the parsing process.
    """
    _token_types.value.remove(token_cls)


def reset_tokens():
    """
    Resets global _token_types to all token classes in __all__.
    """
    global _token_types
    _token_types.value = [globals()[cls_name] for cls_name in __all__]


class BlockToken(object):
    """
    Base class for block-level tokens. Recursively parse inner tokens.

    Naming conventions:

        * lines denotes a list of (possibly unparsed) input lines, and is
          commonly used as the argument name for constructors.

        * BlockToken.children is a list with all the inner tokens (thus if
          a token has children attribute, it is not a leaf node; if a token
          calls span_token.tokenize_inner, it is the boundary between
          span-level tokens and block-level tokens);

        * BlockToken.start takes a line from the document as argument, and
          returns a boolean representing whether that line marks the start
          of the current token. Every subclass of BlockToken must define a
          start function (see block_tokenizer.tokenize).

        * BlockToken.read takes the rest of the lines in the ducment as an
          iterator (including the start line), and consumes all the lines
          that should be read into this token.

          Default to stop at an empty line.

          Note that BlockToken.read does not have to return a list of lines.
          Because the return value of this function will be directly
          passed into the token constructor, we can return any relevant
          parsing information, sometimes even ready-made tokens,
          into the constructor. See block_tokenizer.tokenize.

          If BlockToken.read returns None, the read result is ignored,
          but the token class is responsible for resetting the iterator
          to a previous state. See block_tokenizer.FileWrapper.anchor,
          block_tokenizer.FileWrapper.reset.

    Attributes:
        children (list): inner tokens.
    """

    def __init__(self, lines, tokenize_func):
        self.children = tokenize_func(lines)

    def __contains__(self, text):
        return any(text in child for child in self.children)

    @staticmethod
    def read(lines):
        line_buffer = [next(lines)]
        for line in lines:
            if line == "\n":
                break
            line_buffer.append(line)
        return line_buffer


class Document(BlockToken):
    """
    Document token.
    """

    def __init__(self, lines):
        if isinstance(lines, str):
            lines = lines.splitlines(keepends=True)
        lines = [line if line.endswith("\n") else "{}\n".format(line) for line in lines]
        self.footnotes = {}
        global _root_node
        _root_node = self
        span_token._root_node = self
        self.children = tokenize(lines)
        span_token._root_node = None
        _root_node = None


class Heading(BlockToken):
    """
    Heading token. (["### some heading ###\\n"])
    Boundary between span-level and block-level tokens.

    Attributes:
        level (int): heading level.
        children (list): inner tokens.
    """

    pattern = re.compile(r" {0,3}(#{1,6})(?:\n|\s+?(.*?)(?:\n|\s+?#+\s*?$))")
    level = 0
    content = ""

    def __init__(self, match):
        self.level, content = match
        super().__init__(content, span_token.tokenize_inner)

    @classmethod
    def start(cls, line):
        match_obj = cls.pattern.match(line)
        if match_obj is None:
            return False
        cls.level = len(match_obj.group(1))
        cls.content = (match_obj.group(2) or "").strip()
        if set(cls.content) == {"#"}:
            cls.content = ""
        return True

    @classmethod
    def read(cls, lines):
        next(lines)
        return cls.level, cls.content


class SetextHeading(BlockToken):
    """
    Setext headings.

    Not included in the parsing process, but called by Paragraph.__new__.
    """

    def __init__(self, lines):
        self.level = 1 if lines.pop().lstrip().startswith("=") else 2
        content = "\n".join([line.strip() for line in lines])
        super().__init__(content, span_token.tokenize_inner)

    @classmethod
    def start(cls, line):
        raise NotImplementedError()

    @classmethod
    def read(cls, lines):
        raise NotImplementedError()


class Quote(BlockToken):
    """
    Quote token. (["> # heading\\n", "> paragraph\\n"])
    """

    def __init__(self, parse_buffer):
        # span-level tokenizing happens here.
        self.children = tokenizer.make_tokens(parse_buffer)

    @staticmethod
    def start(line):
        stripped = line.lstrip(" ")
        if len(line) - len(stripped) > 3:
            return False
        return stripped.startswith(">")

    @classmethod
    def read(cls, lines):
        # first line
        line = cls.convert_leading_tabs(next(lines).lstrip()).split(">", 1)[1]
        if len(line) > 0 and line[0] == " ":
            line = line[1:]
        line_buffer = [line]

        # set booleans
        in_code_fence = CodeFence.start(line)
        in_block_code = BlockCode.start(line)
        blank_line = line.strip() == ""

        # loop
        next_line = lines.peek()
        while (
            next_line is not None
            and next_line.strip() != ""
            and not Heading.start(next_line)
            and not CodeFence.start(next_line)
            and not ThematicBreak.start(next_line)
            and not List.start(next_line)
        ):
            stripped = cls.convert_leading_tabs(next_line.lstrip())
            prepend = 0
            if stripped[0] == ">":
                # has leader, not lazy continuation
                prepend += 1
                if stripped[1] == " ":
                    prepend += 1
                stripped = stripped[prepend:]
                in_code_fence = CodeFence.start(stripped)
                in_block_code = BlockCode.start(stripped)
                blank_line = stripped.strip() == ""
                line_buffer.append(stripped)
            elif in_code_fence or in_block_code or blank_line:
                # not paragraph continuation text
                break
            else:
                # lazy continuation, preserve whitespace
                line_buffer.append(next_line)
            next(lines)
            next_line = lines.peek()

        # block level tokens are parsed here, so that footnotes
        # in quotes can be recognized before span-level tokenizing.
        Paragraph.parse_setext = False
        parse_buffer = tokenizer.tokenize_block(line_buffer, _token_types.value)
        Paragraph.parse_setext = True
        return parse_buffer

    @staticmethod
    def convert_leading_tabs(string):
        string = string.replace(">\t", "   ", 1)
        count = 0
        for i, c in enumerate(string):
            if c == "\t":
                count += 4
            elif c == " ":
                count += 1
            else:
                break
        if i == 0:
            return string
        return ">" + " " * count + string[i:]


class Paragraph(BlockToken):
    """
    Paragraph token. (["some\\n", "continuous\\n", "lines\\n"])
    Boundary between span-level and block-level tokens.
    """

    setext_pattern = re.compile(r" {0,3}(=|-)+ *$")
    parse_setext = True  # can be disabled by Quote

    def __new__(cls, lines):
        if not isinstance(lines, list):
            # setext heading token, return directly
            return lines
        return super().__new__(cls)

    def __init__(self, lines):
        content = "".join([line.lstrip() for line in lines]).strip()
        super().__init__(content, span_token.tokenize_inner)

    @staticmethod
    def start(line):
        return line.strip() != ""

    @classmethod
    def read(cls, lines):
        line_buffer = [next(lines)]
        next_line = lines.peek()
        while (
            next_line is not None
            and next_line.strip() != ""
            and not Heading.start(next_line)
            and not CodeFence.start(next_line)
            and not Quote.start(next_line)
        ):

            # check if next_line starts List
            list_pair = ListItem.parse_marker(next_line)
            if len(next_line) - len(next_line.lstrip()) < 4 and list_pair is not None:
                prepend, leader = list_pair
                # non-empty list item
                if next_line[:prepend].endswith(" "):
                    # unordered list, or ordered list starting from 1
                    if not leader[:-1].isdigit() or leader[:-1] == "1":
                        break

            # check if next_line starts HTMLBlock other than type 7
            html_block = HTMLBlock.start(next_line)
            if html_block and html_block != 7:
                break

            # check if we see a setext underline
            if cls.parse_setext and cls.is_setext_heading(next_line):
                line_buffer.append(next(lines))
                return SetextHeading(line_buffer)

            # check if we have a ThematicBreak (has to be after setext)
            if ThematicBreak.start(next_line):
                break

            # no other tokens, we're good
            line_buffer.append(next(lines))
            next_line = lines.peek()
        return line_buffer

    @classmethod
    def is_setext_heading(cls, line):
        return cls.setext_pattern.match(line)


class BlockCode(BlockToken):
    """
    Indented code.

    Attributes:
        children (list): contains a single span_token.RawText token.
        language (str): always the empty string.
    """

    def __init__(self, lines):
        self.language = ""
        self.children = (span_token.RawText("".join(lines).strip("\n") + "\n"),)

    @staticmethod
    def start(line):
        return line.replace("\t", "    ", 1).startswith("    ")

    @classmethod
    def read(cls, lines):
        line_buffer = []
        for line in lines:
            if line.strip() == "":
                line_buffer.append(line.lstrip(" ") if len(line) < 5 else line[4:])
                continue
            if not line.replace("\t", "    ", 1).startswith("    "):
                lines.backstep()
                break
            line_buffer.append(cls.strip(line))
        return line_buffer

    @staticmethod
    def strip(string):
        count = 0
        for i, c in enumerate(string):
            if c == "\t":
                return string[i + 1 :]
            elif c == " ":
                count += 1
            else:
                break
            if count == 4:
                return string[i + 1 :]
        return string


class CodeFence(BlockToken):
    """
    Code fence. (["```sh\\n", "rm -rf /", ..., "```"])
    Boundary between span-level and block-level tokens.

    Attributes:
        children (list): contains a single span_token.RawText token.
        language (str): language of code block (default to empty).
    """

    pattern = re.compile(r"( {0,3})((?:`|~){3,}) *(\S*)")
    _open_info = None

    def __init__(self, match):
        lines, open_info = match
        self.language = span_token.EscapeSequence.strip(open_info[2])
        self.children = (span_token.RawText("".join(lines)),)

    @classmethod
    def start(cls, line):
        match_obj = cls.pattern.match(line)
        if not match_obj:
            return False
        prepend, leader, lang = match_obj.groups()
        if leader[0] in lang or leader[0] in line[match_obj.end() :]:
            return False
        cls._open_info = len(prepend), leader, lang
        return True

    @classmethod
    def read(cls, lines):
        next(lines)
        line_buffer = []
        for line in lines:
            stripped_line = line.lstrip(" ")
            diff = len(line) - len(stripped_line)
            if (
                stripped_line.startswith(cls._open_info[1])
                and len(stripped_line.split(maxsplit=1)) == 1
                and diff < 4
            ):
                break
            if diff > cls._open_info[0]:
                stripped_line = " " * (diff - cls._open_info[0]) + stripped_line
            line_buffer.append(stripped_line)
        return line_buffer, cls._open_info


class List(BlockToken):
    """
    List token.

    Attributes:
        children (list): a list of ListItem tokens.
        loose (bool): whether the list is loose.
        start (NoneType or int): None if unordered, starting number if ordered.
    """

    pattern = re.compile(r" {0,3}(?:\d{0,9}[.)]|[+\-*])(?:[ \t]*$|[ \t]+)")

    def __init__(self, matches):
        self.children = [ListItem(*match) for match in matches]
        self.loose = any(item.loose for item in self.children)
        leader = self.children[0].leader
        self.start = None
        if len(leader) != 1:
            self.start = int(leader[:-1])

    @classmethod
    def start(cls, line):
        return cls.pattern.match(line)

    @classmethod
    def read(cls, lines):
        leader = None
        next_marker = None
        matches = []
        while True:
            output, next_marker = ListItem.read(lines, next_marker)
            item_leader = output[2]
            if leader is None:
                leader = item_leader
            elif not cls.same_marker_type(leader, item_leader):
                lines.reset()
                break
            matches.append(output)
            if next_marker is None:
                break

        if matches:
            # Only consider the last list item loose if there's more than one element
            last_parse_buffer = matches[-1][0]
            last_parse_buffer.loose = (
                len(last_parse_buffer) > 1 and last_parse_buffer.loose
            )

        return matches

    @staticmethod
    def same_marker_type(leader, other):
        if len(leader) == 1:
            return leader == other
        return (
            leader[:-1].isdigit() and other[:-1].isdigit() and leader[-1] == other[-1]
        )


class ListItem(BlockToken):
    """
    List items. Not included in the parsing process, but called by List.
    """

    pattern = re.compile(r"\s*(\d{0,9}[.)]|[+\-*])(\s*$|\s+)")

    def __init__(self, parse_buffer, prepend, leader):
        self.leader = leader
        self.prepend = prepend
        self.children = tokenizer.make_tokens(parse_buffer)
        self.loose = parse_buffer.loose

    @staticmethod
    def in_continuation(line, prepend):
        return line.strip() == "" or len(line) - len(line.lstrip()) >= prepend

    @staticmethod
    def other_token(line):
        return (
            Heading.start(line)
            or Quote.start(line)
            or CodeFence.start(line)
            or ThematicBreak.start(line)
        )

    @classmethod
    def parse_marker(cls, line):
        """
        Returns a pair (prepend, leader) iff the line has a valid leader.
        """
        match_obj = cls.pattern.match(line)
        if match_obj is None:
            return None  # no valid leader
        leader = match_obj.group(1)
        content = match_obj.group(0).replace(leader + "\t", leader + "   ", 1)
        # reassign prepend and leader
        prepend = len(content)
        if prepend == len(line.rstrip("\n")):
            prepend = match_obj.end(1) + 1
        else:
            spaces = match_obj.group(2)
            if spaces.startswith("\t"):
                spaces = spaces.replace("\t", "   ", 1)
            spaces = spaces.replace("\t", "    ")
            n_spaces = len(spaces)
            if n_spaces > 4:
                prepend = match_obj.end(1) + 1
        return prepend, leader

    @classmethod
    def read(cls, lines, prev_marker=None):
        next_marker = None
        lines.anchor()
        prepend = -1
        leader = None
        line_buffer = []

        # first line
        line = next(lines)
        prepend, leader = prev_marker if prev_marker else cls.parse_marker(line)
        line = line.replace(leader + "\t", leader + "   ", 1).replace("\t", "    ")
        empty_first_line = line[prepend:].strip() == ""
        if not empty_first_line:
            line_buffer.append(line[prepend:])
        next_line = lines.peek()
        if empty_first_line and next_line is not None and next_line.strip() == "":
            parse_buffer = tokenizer.tokenize_block([next(lines)], _token_types.value)
            next_line = lines.peek()
            if next_line is not None:
                marker_info = cls.parse_marker(next_line)
                if marker_info is not None:
                    next_marker = marker_info
            return (parse_buffer, prepend, leader), next_marker

        # loop
        newline = 0
        while True:
            # no more lines
            if next_line is None:
                # strip off newlines
                if newline:
                    lines.backstep()
                    del line_buffer[-newline:]
                break
            next_line = next_line.replace("\t", "    ")
            # not in continuation
            if not cls.in_continuation(next_line, prepend):
                # directly followed by another token
                if cls.other_token(next_line):
                    if newline:
                        lines.backstep()
                        del line_buffer[-newline:]
                    break
                # next_line is a new list item
                marker_info = cls.parse_marker(next_line)
                if marker_info is not None:
                    next_marker = marker_info
                    break
                # not another item, has newlines -> not continuation
                if newline:
                    lines.backstep()
                    del line_buffer[-newline:]
                    break
            next(lines)
            line = next_line
            stripped = line.lstrip(" ")
            diff = len(line) - len(stripped)
            if diff > prepend:
                stripped = " " * (diff - prepend) + stripped
            line_buffer.append(stripped)
            newline = newline + 1 if next_line.strip() == "" else 0
            next_line = lines.peek()

        # block-level tokens are parsed here, so that footnotes can be
        # recognized before span-level parsing.
        parse_buffer = tokenizer.tokenize_block(line_buffer, _token_types.value)
        return (parse_buffer, prepend, leader), next_marker


class Table(BlockToken):
    """
    Table token.

    Attributes:
        has_header (bool): whether table has header row.
        column_align (list): align options for each column (default to [None]).
        children (list): inner tokens (TableRows).
    """

    def __init__(self, lines):
        if "---" in lines[1]:
            self.column_align = [
                self.parse_align(column) for column in self.split_delimiter(lines[1])
            ]
            self.header = TableRow(lines[0], self.column_align)
            self.children = [TableRow(line, self.column_align) for line in lines[2:]]
        else:
            self.column_align = [None]
            self.children = [TableRow(line) for line in lines]

    @staticmethod
    def split_delimiter(delimiter):
        """
        Helper function; returns a list of align options.

        Args:
            delimiter (str): e.g.: "| :--- | :---: | ---: |\n"

        Returns:
            a list of align options (None, 0 or 1).
        """
        return re.findall(r":?---+:?", delimiter)

    @staticmethod
    def parse_align(column):
        """
        Helper function; returns align option from cell content.

        Returns:
            None if align = left;
            0    if align = center;
            1    if align = right.
        """
        return (0 if column[0] == ":" else 1) if column[-1] == ":" else None

    @staticmethod
    def start(line):
        return "|" in line

    @staticmethod
    def read(lines):
        lines.anchor()
        line_buffer = [next(lines)]
        while lines.peek() is not None and "|" in lines.peek():
            line_buffer.append(next(lines))
        if len(line_buffer) < 2 or "---" not in line_buffer[1]:
            lines.reset()
            return None
        return line_buffer


class TableRow(BlockToken):
    """
    Table row token.

    Should only be called by Table.__init__().
    """

    def __init__(self, line, row_align=None):
        self.row_align = row_align or [None]
        cells = filter(None, line.strip().split("|"))
        self.children = [
            TableCell(cell.strip() if cell else "", align)
            for cell, align in zip_longest(cells, self.row_align)
        ]


class TableCell(BlockToken):
    """
    Table cell token.
    Boundary between span-level and block-level tokens.

    Should only be called by TableRow.__init__().

    Attributes:
        align (bool): align option for current cell (default to None).
        children (list): inner (span-)tokens.
    """

    def __init__(self, content, align=None):
        self.align = align
        super().__init__(content, span_token.tokenize_inner)


class Footnote(BlockToken):
    """
    Footnote token.

    The constructor returns None, because the footnote information
    is stored in Footnote.read.
    """

    label_pattern = re.compile(r"[ \n]{0,3}\[(.+?)\]", re.DOTALL)

    def __new__(cls, _):
        return None

    @classmethod
    def start(cls, line):
        return line.lstrip().startswith("[")

    @classmethod
    def read(cls, lines):
        line_buffer = []
        next_line = lines.peek()
        while next_line is not None and next_line.strip() != "":
            line_buffer.append(next(lines))
            next_line = lines.peek()
        string = "".join(line_buffer)
        offset = 0
        matches = []
        while offset < len(string) - 1:
            match_info = cls.match_reference(lines, string, offset)
            if match_info is None:
                break
            offset, match = match_info
            matches.append(match)
        cls.append_footnotes(matches, _root_node)
        return matches or None

    @classmethod
    def match_reference(cls, lines, string, offset):
        match_info = cls.match_link_label(string, offset)
        if not match_info:
            cls.backtrack(lines, string, offset)
            return None
        _, label_end, label = match_info

        if not follows(string, label_end - 1, ":"):
            cls.backtrack(lines, string, offset)
            return None

        match_info = cls.match_link_dest(string, label_end)
        if not match_info:
            cls.backtrack(lines, string, offset)
            return None
        _, dest_end, dest = match_info

        match_info = cls.match_link_title(string, dest_end)
        if not match_info:
            cls.backtrack(lines, string, dest_end)
            return None
        _, title_end, title = match_info

        return title_end, (label, dest, title)

    @classmethod
    def match_link_label(cls, string, offset):
        start = -1
        end = -1
        escaped = False
        for i, c in enumerate(string[offset:], start=offset):
            if c == "\\" and not escaped:
                escaped = True
            elif c == "[" and not escaped:
                if start == -1:
                    start = i
                else:
                    return None
            elif c == "]" and not escaped:
                end = i
                label = string[start + 1 : end]
                if label.strip() != "":
                    return start, end + 1, label
                return None
            elif escaped:
                escaped = False
        return None

    @classmethod
    def match_link_dest(cls, string, offset):
        offset = shift_whitespace(string, offset + 1)
        if offset == len(string):
            return None
        if string[offset] == "<":
            escaped = False
            for i, c in enumerate(string[offset + 1 :], start=offset + 1):
                if c == "\\" and not escaped:
                    escaped = True
                elif c == " " or c == "\n" or (c == "<" and not escaped):
                    return None
                elif c == ">" and not escaped:
                    return offset, i + 1, string[offset + 1 : i]
                elif escaped:
                    escaped = False
            return None
        else:
            escaped = False
            count = 0
            for i, c in enumerate(string[offset:], start=offset):
                if c == "\\" and not escaped:
                    escaped = True
                elif c in whitespace:
                    break
                elif not escaped:
                    if c == "(":
                        count += 1
                    elif c == ")":
                        count -= 1
                elif is_control_char(c):
                    return None
                elif escaped:
                    escaped = False
            if count != 0:
                return None
            return offset, i, string[offset:i]

    @classmethod
    def match_link_title(cls, string, offset):
        new_offset = shift_whitespace(string, offset)
        if (
            new_offset == len(string)
            or "\n" in string[offset:new_offset]
            and string[new_offset] == "["
        ):
            return offset, new_offset, ""
        if string[new_offset] == '"':
            closing = '"'
        elif string[new_offset] == "'":
            closing = "'"
        elif string[new_offset] == "(":
            closing = ")"
        elif "\n" in string[offset:new_offset]:
            return offset, offset, ""
        else:
            return None
        offset = new_offset
        escaped = False
        for i, c in enumerate(string[offset + 1 :], start=offset + 1):
            if c == "\\" and not escaped:
                escaped = True
            elif c == closing and not escaped:
                new_offset = shift_whitespace(string, i + 1)
                if "\n" not in string[i + 1 : new_offset]:
                    return None
                return offset, new_offset, string[offset + 1 : i]
            elif escaped:
                escaped = False
        return None

    @staticmethod
    def append_footnotes(matches, root):
        for key, dest, title in matches:
            key = normalize_label(key)
            dest = span_token.EscapeSequence.strip(dest.strip())
            title = span_token.EscapeSequence.strip(title)
            if key not in root.footnotes:
                root.footnotes[key] = dest, title

    @staticmethod
    def backtrack(lines, string, offset):
        lines._index -= string[offset + 1 :].count("\n")


class ThematicBreak(BlockToken):
    """
    Thematic break token (a.k.a. horizontal rule.)
    """

    pattern = re.compile(r" {0,3}(?:([-_*])\s*?)(?:\1\s*?){2,}$")

    def __init__(self, _):
        pass

    @classmethod
    def start(cls, line):
        return cls.pattern.match(line)

    @staticmethod
    def read(lines):
        return [next(lines)]


class HTMLBlock(BlockToken):
    """
    Block-level HTML tokens.

    Attributes:
        content (str): literal strings rendered as-is.
    """

    _end_cond = None
    multiblock = re.compile(r"<(script|pre|style)[ >\n]")
    predefined = re.compile(r"<\/?(.+?)(?:\/?>|[ \n])")
    custom_tag = re.compile(
        r"(?:" + "|".join((span_token._open_tag, span_token._closing_tag)) + r")\s*$"
    )

    def __init__(self, lines):
        self.content = "".join(lines).rstrip("\n")

    @classmethod
    def start(cls, line):
        stripped = line.lstrip()
        if len(line) - len(stripped) >= 4:
            return False
        # rule 1: <pre>, <script> or <style> tags, allow newlines in block
        match_obj = cls.multiblock.match(stripped)
        if match_obj is not None:
            cls._end_cond = "</{}>".format(match_obj.group(1).casefold())
            return 1
        # rule 2: html comment tags, allow newlines in block
        if stripped.startswith("<!--"):
            cls._end_cond = "-->"
            return 2
        # rule 3: tags that starts with <?, allow newlines in block
        if stripped.startswith("<?"):
            cls._end_cond = "?>"
            return 3
        # rule 4: tags that starts with <!, allow newlines in block
        if stripped.startswith("<!") and stripped[2].isupper():
            cls._end_cond = ">"
            return 4
        # rule 5: CDATA declaration, allow newlines in block
        if stripped.startswith("<![CDATA["):
            cls._end_cond = "]]>"
            return 5
        # rule 6: predefined tags (see html_token._tags), read until newline
        match_obj = cls.predefined.match(stripped)
        if match_obj is not None and match_obj.group(1).casefold() in span_token._tags:
            cls._end_cond = None
            return 6
        # rule 7: custom tags, read until newline
        match_obj = cls.custom_tag.match(stripped)
        if match_obj is not None:
            cls._end_cond = None
            return 7
        return False

    @classmethod
    def read(cls, lines):
        # note: stop condition can trigger on the starting line
        line_buffer = []
        for line in lines:
            line_buffer.append(line)
            if cls._end_cond is not None:
                if cls._end_cond in line.casefold():
                    break
            elif line.strip() == "":
                line_buffer.pop()
                break
        return line_buffer


_token_types = local()
_token_types.value = []
reset_tokens()



import unittest
from mistletoe import Document, ast_renderer


class TestASTRenderer(unittest.TestCase):
    def test(self):
        self.maxDiff = None
        d = Document(["# heading 1\n", "\n", "hello\n", "world\n"])
        output = ast_renderer.get_ast(d)
        target = {
            "type": "Document",
            "footnotes": {},
            "children": [
                {
                    "type": "Heading",
                    "level": 1,
                    "children": [{"type": "RawText", "content": "heading 1"}],
                },
                {
                    "type": "Paragraph",
                    "children": [
                        {"type": "RawText", "content": "hello"},
                        {"type": "LineBreak", "soft": True, "content": ""},
                        {"type": "RawText", "content": "world"},
                    ],
                },
            ],
        }
        self.assertEqual(output, target)

    def test_footnotes(self):
        self.maxDiff = None
        d = Document(["[bar][baz]\n", "\n", "[baz]: spam\n"])
        target = {
            "type": "Document",
            "footnotes": {"baz": ("spam", "")},
            "children": [
                {
                    "type": "Paragraph",
                    "children": [
                        {
                            "type": "Link",
                            "target": "spam",
                            "title": "",
                            "children": [{"type": "RawText", "content": "bar"}],
                        }
                    ],
                }
            ],
        }
        output = ast_renderer.get_ast(d)
        self.assertEqual(output, target)



from unittest import TestCase, mock
from mistletoe.html_renderer import HTMLRenderer


class TestRenderer(TestCase):
    def setUp(self):
        self.renderer = HTMLRenderer()
        self.renderer.render_inner = mock.Mock(return_value="inner")
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def _test_token(
        self, token_name, output, children=True, without_attrs=None, **kwargs
    ):
        render_func = self.renderer.render_map[token_name]
        children = mock.MagicMock(spec=list) if children else None
        mock_token = mock.Mock(children=children, **kwargs)
        without_attrs = without_attrs or []
        for attr in without_attrs:
            delattr(mock_token, attr)
        self.assertEqual(render_func(mock_token), output)


class TestHTMLRenderer(TestRenderer):
    def test_strong(self):
        self._test_token("Strong", "<strong>inner</strong>")

    def test_emphasis(self):
        self._test_token("Emphasis", "<em>inner</em>")

    def test_inline_code(self):
        from mistletoe.span_token import tokenize_inner

        rendered = self.renderer.render(tokenize_inner("`foo`")[0])
        self.assertEqual(rendered, "<code>foo</code>")

    def test_strikethrough(self):
        self._test_token("Strikethrough", "<del>inner</del>")

    def test_image(self):
        output = '<img src="src" alt="" title="title" />'
        self._test_token("Image", output, src="src", title="title")

    def test_link(self):
        output = '<a href="target" title="title">inner</a>'
        self._test_token("Link", output, target="target", title="title")

    def test_autolink(self):
        output = '<a href="link">inner</a>'
        self._test_token("AutoLink", output, target="link", mailto=False)

    def test_escape_sequence(self):
        self._test_token("EscapeSequence", "inner")

    def test_raw_text(self):
        self._test_token(
            "RawText", "john &amp; jane", children=False, content="john & jane"
        )

    def test_html_span(self):
        self._test_token(
            "HTMLSpan", "<some>text</some>", children=False, content="<some>text</some>"
        )

    def test_heading(self):
        output = "<h3>inner</h3>"
        self._test_token("Heading", output, level=3)

    def test_quote(self):
        output = "<blockquote>\n</blockquote>"
        self._test_token("Quote", output)

    def test_paragraph(self):
        self._test_token("Paragraph", "<p>inner</p>")

    def test_block_code(self):
        from mistletoe.block_token import tokenize

        rendered = self.renderer.render(tokenize(["```sh\n", "foo\n", "```\n"])[0])
        output = '<pre><code class="language-sh">foo\n</code></pre>'
        self.assertEqual(rendered, output)

    def test_block_code_no_language(self):
        from mistletoe.block_token import tokenize

        rendered = self.renderer.render(tokenize(["```\n", "foo\n", "```\n"])[0])
        output = "<pre><code>foo\n</code></pre>"
        self.assertEqual(rendered, output)

    def test_list(self):
        output = "<ul>\n\n</ul>"
        self._test_token("List", output, start=None)

    def test_list_item(self):
        output = "<li></li>"
        self._test_token("ListItem", output)

    def test_table_with_header(self):
        func_path = "mistletoe.html_renderer.HTMLRenderer.render_table_row"
        with mock.patch(func_path, autospec=True) as mock_func:
            mock_func.return_value = "row"
            output = (
                "<table>\n"
                "<thead>\nrow</thead>\n"
                "<tbody>\ninner</tbody>\n"
                "</table>"
            )
            self._test_token("Table", output)

    def test_table_without_header(self):
        func_path = "mistletoe.html_renderer.HTMLRenderer.render_table_row"
        with mock.patch(func_path, autospec=True) as mock_func:
            mock_func.return_value = "row"
            output = "<table>\n<tbody>\ninner</tbody>\n</table>"
            self._test_token("Table", output, without_attrs=["header"])

    def test_table_row(self):
        self._test_token("TableRow", "<tr>\n</tr>\n")

    def test_table_cell(self):
        output = '<td align="left">inner</td>\n'
        self._test_token("TableCell", output, align=None)

    def test_table_cell0(self):
        output = '<td align="center">inner</td>\n'
        self._test_token("TableCell", output, align=0)

    def test_table_cell1(self):
        output = '<td align="right">inner</td>\n'
        self._test_token("TableCell", output, align=1)

    def test_thematic_break(self):
        self._test_token("ThematicBreak", "<hr />", children=False)

    def test_html_block(self):
        content = output = "<h1>hello</h1>\n<p>this is\na paragraph</p>\n"
        self._test_token("HTMLBlock", output, children=False, content=content)

    def test_line_break(self):
        self._test_token("LineBreak", "<br />\n", children=False, soft=False)

    def test_document(self):
        self._test_token("Document", "", footnotes={})


class TestHTMLRendererFootnotes(TestCase):
    def setUp(self):
        self.renderer = HTMLRenderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def test_footnote_image(self):
        from mistletoe import Document

        token = Document(["![alt][foo]\n", "\n", '[foo]: bar "title"\n'])
        output = '<p><img src="bar" alt="alt" title="title" /></p>\n'
        self.assertEqual(self.renderer.render(token), output)

    def test_footnote_link(self):
        from mistletoe import Document

        token = Document(["[name][foo]\n", "\n", "[foo]: target\n"])
        output = '<p><a href="target">name</a></p>\n'
        self.assertEqual(self.renderer.render(token), output)



import unittest
from unittest.mock import patch, call
from mistletoe import block_token, span_token
from mistletoe.block_tokenizer import FileWrapper


class TestToken(unittest.TestCase):
    def setUp(self):
        self.addCleanup(
            lambda: span_token._token_types.value.__setitem__(-1, span_token.RawText)
        )
        patcher = patch("mistletoe.span_token.RawText")
        self.mock = patcher.start()
        span_token._token_types.value[-1] = self.mock
        self.addCleanup(patcher.stop)

    def _test_match(self, token_cls, lines, arg, **kwargs):
        token = next(iter(block_token.tokenize(lines)))
        self.assertIsInstance(token, token_cls)
        self._test_token(token, arg, **kwargs)

    def _test_token(self, token, arg, **kwargs):
        for attr, value in kwargs.items():
            self.assertEqual(getattr(token, attr), value)
        self.mock.assert_any_call(arg)


class TestATXHeading(TestToken):
    def test_match(self):
        lines = ["### heading 3\n"]
        arg = "heading 3"
        self._test_match(block_token.Heading, lines, arg, level=3)

    def test_children_with_enclosing_hashes(self):
        lines = ["# heading 3 #####  \n"]
        arg = "heading 3"
        self._test_match(block_token.Heading, lines, arg, level=1)

    def test_not_heading(self):
        lines = ["####### paragraph\n"]
        arg = "####### paragraph"
        self._test_match(block_token.Paragraph, lines, arg)

    def test_heading_in_paragraph(self):
        lines = ["foo\n", "# heading\n", "bar\n"]
        token1, token2, token3 = block_token.tokenize(lines)
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertIsInstance(token2, block_token.Heading)
        self.assertIsInstance(token3, block_token.Paragraph)


class TestSetextHeading(TestToken):
    def test_match(self):
        lines = ["some heading\n", "---\n"]
        arg = "some heading"
        self._test_match(block_token.SetextHeading, lines, arg, level=2)

    def test_next(self):
        lines = ["some\n", "heading\n", "---\n", "\n", "foobar\n"]
        tokens = iter(block_token.tokenize(lines))
        self.assertIsInstance(next(tokens), block_token.SetextHeading)
        self.assertIsInstance(next(tokens), block_token.Paragraph)
        self.mock.assert_has_calls([call("some"), call("heading"), call("foobar")])
        with self.assertRaises(StopIteration):
            next(tokens)


class TestQuote(unittest.TestCase):
    def test_match(self):
        with patch("mistletoe.block_token.Paragraph"):
            token = next(iter(block_token.tokenize(["> line 1\n", "> line 2\n"])))
            self.assertIsInstance(token, block_token.Quote)

    def test_lazy_continuation(self):
        with patch("mistletoe.block_token.Paragraph"):
            token = next(iter(block_token.tokenize(["> line 1\n", "line 2\n"])))
            self.assertIsInstance(token, block_token.Quote)


class TestCodeFence(TestToken):
    def test_match_fenced_code(self):
        lines = ["```sh\n", "rm dir\n", "mkdir test\n", "```\n"]
        arg = "rm dir\nmkdir test\n"
        self._test_match(block_token.CodeFence, lines, arg, language="sh")

    def test_match_fenced_code_with_tilda(self):
        lines = ["~~~sh\n", "rm dir\n", "mkdir test\n", "~~~\n"]
        arg = "rm dir\nmkdir test\n"
        self._test_match(block_token.CodeFence, lines, arg, language="sh")

    def test_mixed_code_fence(self):
        lines = ["~~~markdown\n", "```sh\n", "some code\n", "```\n", "~~~\n"]
        arg = "```sh\nsome code\n```\n"
        self._test_match(block_token.CodeFence, lines, arg, language="markdown")

    def test_fence_code_lazy_continuation(self):
        lines = ["```sh\n", "rm dir\n", "\n", "mkdir test\n", "```\n"]
        arg = "rm dir\n\nmkdir test\n"
        self._test_match(block_token.CodeFence, lines, arg, language="sh")

    def test_no_wrapping_newlines_code_fence(self):
        lines = ["```\n", "hey", "```\n", "paragraph\n"]
        arg = "hey"
        self._test_match(block_token.CodeFence, lines, arg, language="")

    def test_unclosed_code_fence(self):
        lines = ["```\n", "hey"]
        arg = "hey"
        self._test_match(block_token.CodeFence, lines, arg, language="")


class TestBlockCode(TestToken):
    def test_parse_indented_code(self):
        lines = ["    rm dir\n", "    mkdir test\n"]
        arg = "rm dir\nmkdir test\n"
        self._test_match(block_token.BlockCode, lines, arg, language="")


class TestParagraph(TestToken):
    def test_parse(self):
        lines = ["some\n", "continuous\n", "lines\n"]
        arg = "some"
        self._test_match(block_token.Paragraph, lines, arg)

    def test_read(self):
        lines = ["this\n", "```\n", "is some\n", "```\n", "code\n"]
        try:
            token1, token2, token3 = block_token.tokenize(lines)
        except ValueError as e:
            raise AssertionError("Token number mismatch.") from e
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertIsInstance(token2, block_token.CodeFence)
        self.assertIsInstance(token3, block_token.Paragraph)


class TestListItem(unittest.TestCase):
    def test_parse_marker(self):
        lines = [
            "- foo\n",
            "*    bar\n",
            " + baz\n",
            "1. item 1\n",
            "2) item 2\n",
            "123456789. item x\n",
        ]
        for line in lines:
            self.assertTrue(block_token.ListItem.parse_marker(line))
        bad_lines = ["> foo\n", "1item 1\n", "2| item 2\n", "1234567890. item x\n"]
        for line in bad_lines:
            self.assertFalse(block_token.ListItem.parse_marker(line))

    def test_tokenize(self):
        lines = [" -    foo\n", "   bar\n", "\n", "          baz\n"]
        token1, token2 = next(iter(block_token.tokenize(lines))).children[0].children
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertTrue("foo" in token1)
        self.assertIsInstance(token2, block_token.BlockCode)

    def test_sublist(self):
        lines = ["- foo\n", "  - bar\n"]
        token1, token2 = block_token.tokenize(lines)[0].children[0].children
        self.assertIsInstance(token1, block_token.Paragraph)
        self.assertIsInstance(token2, block_token.List)

    def test_deep_list(self):
        lines = ["- foo\n", "  - bar\n", "    - baz\n"]
        FileWrapper(lines)
        ptoken, ltoken = block_token.tokenize(lines)[0].children[0].children
        self.assertIsInstance(ptoken, block_token.Paragraph)
        self.assertIsInstance(ltoken, block_token.List)
        self.assertTrue("foo" in ptoken)
        ptoken, ltoken = ltoken.children[0].children
        self.assertIsInstance(ptoken, block_token.Paragraph)
        self.assertTrue("bar" in ptoken)
        self.assertIsInstance(ltoken, block_token.List)
        self.assertTrue("baz" in ltoken)

    def test_loose_list(self):
        lines = ["- foo\n", "  ~~~\n", "  bar\n", "  \n", "  baz\n" "  ~~~\n"]
        FileWrapper(lines)
        list_item = block_token.tokenize(lines)[0].children[0]
        self.assertEqual(list_item.loose, False)

    def test_tight_list(self):
        lines = ["- foo\n", "\n", "# bar\n"]
        FileWrapper(lines)
        list_item = block_token.tokenize(lines)[0].children[0]
        self.assertEqual(list_item.loose, False)


class TestList(unittest.TestCase):
    def test_different_markers(self):
        lines = ["- foo\n", "* bar\n", "1. baz\n", "2) spam\n"]
        l1, l2, l3, l4 = block_token.tokenize(lines)
        self.assertIsInstance(l1, block_token.List)
        self.assertTrue("foo" in l1)
        self.assertIsInstance(l2, block_token.List)
        self.assertTrue("bar" in l2)
        self.assertIsInstance(l3, block_token.List)
        self.assertTrue("baz" in l3)
        self.assertIsInstance(l4, block_token.List)
        self.assertTrue("spam" in l4)

    def test_sublist(self):
        lines = ["- foo\n", "  + bar\n"]
        (token,) = block_token.tokenize(lines)
        self.assertIsInstance(token, block_token.List)


class TestTable(unittest.TestCase):
    def test_parse_align(self):
        test_func = block_token.Table.parse_align
        self.assertEqual(test_func(":------"), None)
        self.assertEqual(test_func(":-----:"), 0)
        self.assertEqual(test_func("------:"), 1)

    def test_parse_delimiter(self):
        test_func = block_token.Table.split_delimiter
        self.assertEqual(
            list(test_func("| :--- | :---: | ---:|\n")), [":---", ":---:", "---:"]
        )

    def test_match(self):
        lines = [
            "| header 1 | header 2 | header 3 |\n",
            "| --- | --- | --- |\n",
            "| cell 1 | cell 2 | cell 3 |\n",
            "| more 1 | more 2 | more 3 |\n",
        ]
        with patch("mistletoe.block_token.TableRow") as mock:
            token = next(iter(block_token.tokenize(lines)))
            self.assertIsInstance(token, block_token.Table)
            self.assertTrue(hasattr(token, "header"))
            self.assertEqual(token.column_align, [None, None, None])
            token.children
            calls = [call(line, [None, None, None]) for line in lines[:1] + lines[2:]]
            mock.assert_has_calls(calls)

    def test_easy_table(self):
        lines = ["header 1 | header 2\n", "    ---: | :---\n", "  cell 1 | cell 2\n"]
        with patch("mistletoe.block_token.TableRow") as mock:
            (token,) = block_token.tokenize(lines)
            self.assertIsInstance(token, block_token.Table)
            self.assertTrue(hasattr(token, "header"))
            self.assertEqual(token.column_align, [1, None])
            token.children
            calls = [call(line, [1, None]) for line in lines[:1] + lines[2:]]
            mock.assert_has_calls(calls)

    def test_not_easy_table(self):
        lines = ["not header 1 | not header 2\n", "foo | bar\n"]
        (token,) = block_token.tokenize(lines)
        self.assertIsInstance(token, block_token.Paragraph)


class TestTableRow(unittest.TestCase):
    def test_match(self):
        with patch("mistletoe.block_token.TableCell") as mock:
            line = "| cell 1 | cell 2 |\n"
            token = block_token.TableRow(line)
            self.assertEqual(token.row_align, [None])
            token.children
            mock.assert_has_calls([call("cell 1", None), call("cell 2", None)])

    def test_easy_table_row(self):
        with patch("mistletoe.block_token.TableCell") as mock:
            line = "cell 1 | cell 2\n"
            token = block_token.TableRow(line)
            self.assertEqual(token.row_align, [None])
            token.children
            mock.assert_has_calls([call("cell 1", None), call("cell 2", None)])

    def test_short_row(self):
        with patch("mistletoe.block_token.TableCell") as mock:
            line = "| cell 1 |\n"
            token = block_token.TableRow(line, [None, None])
            self.assertEqual(token.row_align, [None, None])
            token.children
            mock.assert_has_calls([call("cell 1", None), call("", None)])


class TestTableCell(TestToken):
    def test_match(self):
        token = block_token.TableCell("cell 2")
        self._test_token(token, "cell 2", align=None)


class TestFootnote(unittest.TestCase):
    def test_store(self):
        lines = ["[key 1]: value1\n", "[key 2]: value2\n"]
        token = block_token.Document(lines)
        self.assertEqual(
            token.footnotes, {"key 1": ("value1", ""), "key 2": ("value2", "")}
        )


class TestDocument(unittest.TestCase):
    def test_store_footnote(self):
        lines = ["[key 1]: value1\n", "[key 2]: value2\n"]
        document = block_token.Document(lines)
        self.assertEqual(document.footnotes["key 1"], ("value1", ""))
        self.assertEqual(document.footnotes["key 2"], ("value2", ""))

    def test_auto_splitlines(self):
        lines = "some\ncontinual\nlines\n"
        document = block_token.Document(lines)
        self.assertIsInstance(document.children[0], block_token.Paragraph)
        self.assertEqual(len(document.children), 1)


class TestThematicBreak(unittest.TestCase):
    def test_match(self):
        def test_case(line):
            token = next(iter(block_token.tokenize([line])))
            self.assertIsInstance(token, block_token.ThematicBreak)

        cases = ["---\n", "* * *\n", "_    _    _\n"]
        for case in cases:
            test_case(case)


class TestContains(unittest.TestCase):
    def test_contains(self):
        lines = ["# heading\n", "\n", "paragraph\n", "with\n", "`code`\n"]
        token = block_token.Document(lines)
        self.assertTrue("heading" in token)
        self.assertTrue("code" in token)
        self.assertFalse("foo" in token)



from unittest import TestCase, mock
from mistletoe.latex_renderer import LaTeXRenderer


class TestLaTeXRenderer(TestCase):
    def setUp(self):
        self.renderer = LaTeXRenderer()
        self.renderer.render_inner = mock.Mock(return_value="inner")
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def _test_token(
        self, token_name, output, children=True, without_attrs=None, **kwargs
    ):
        render_func = self.renderer.render_map[token_name]
        children = mock.MagicMock(spec=list) if children else None
        mock_token = mock.Mock(children=children, **kwargs)
        without_attrs = without_attrs or []
        for attr in without_attrs:
            delattr(mock_token, attr)
        self.assertEqual(render_func(mock_token), output)

    def test_strong(self):
        self._test_token("Strong", "\\textbf{inner}")

    def test_emphasis(self):
        self._test_token("Emphasis", "\\textit{inner}")

    def test_inline_code(self):
        self._test_token("InlineCode", "\\verb|inner|")

    def test_strikethrough(self):
        self._test_token("Strikethrough", "\\sout{inner}")

    def test_image(self):
        output = "\n\\includegraphics{src}\n"
        self._test_token("Image", output, src="src")

    def test_link(self):
        output = "\\href{target}{inner}"
        self._test_token("Link", output, target="target")

    def test_autolink(self):
        output = "\\url{target}"
        self._test_token("AutoLink", output, target="target")

    def test_math(self):
        output = "$ 1 + 2 = 3 $"
        self._test_token("Math", output, children=False, content="$ 1 + 2 = 3 $")

    def test_raw_text(self):
        output = "\\$\\&\\#\\{\\}"
        self._test_token("RawText", output, children=False, content="$&#{}")

    def test_heading(self):
        output = "\n\\section{inner}\n"
        self._test_token("Heading", output, level=1)

    def test_quote(self):
        output = "\\begin{displayquote}\ninner\\end{displayquote}\n"
        self._test_token("Quote", output)

    def test_paragraph(self):
        output = "\ninner\n"
        self._test_token("Paragraph", output)

    def test_block_code(self):
        func_path = "mistletoe.latex_renderer.LaTeXRenderer.render_raw_text"
        with mock.patch(func_path, return_value="inner"):
            output = "\n\\begin{lstlisting}[language=sh]\ninner\\end{lstlisting}\n"
            self._test_token("BlockCode", output, language="sh")

    def test_list(self):
        output = "\\begin{itemize}\ninner\\end{itemize}\n"
        self._test_token("List", output, start=None)

    def test_list_item(self):
        self._test_token("ListItem", "\\item inner\n")

    def test_table_with_header(self):
        func_path = "mistletoe.latex_renderer.LaTeXRenderer.render_table_row"
        with mock.patch(func_path, autospec=True, return_value="row\n"):
            output = "\\begin{tabular}{l c r}\nrow\n\\hline\ninner\\end{tabular}\n"
            self._test_token("Table", output, column_align=[None, 0, 1])

    def test_table_without_header(self):
        output = "\\begin{tabular}\ninner\\end{tabular}\n"
        self._test_token("Table", output, without_attrs=["header"], column_align=[None])

    def test_table_row(self):
        self._test_token("TableRow", " \\\\\n")

    def test_table_cell(self):
        self._test_token("TableCell", "inner")

    def test_thematic_break(self):
        self._test_token("ThematicBreak", "\\hrulefill\n")

    def test_line_break(self):
        self._test_token("LineBreak", "\\newline\n", soft=False)

    def test_document(self):
        output = (
            "\\documentclass{article}\n"
            "\\begin{document}\n"
            "inner"
            "\\end{document}\n"
        )
        self._test_token("Document", output, footnotes={})


class TestLaTeXFootnotes(TestCase):
    def setUp(self):
        self.renderer = LaTeXRenderer()
        self.renderer.__enter__()
        self.addCleanup(self.renderer.__exit__, None, None, None)

    def test_footnote_image(self):
        from mistletoe import Document

        raw = ["![alt][foo]\n", "\n", '[foo]: bar "title"\n']
        target = (
            "\\documentclass{article}\n"
            "\\usepackage{graphicx}\n"
            "\\begin{document}\n"
            "\n"
            "\n\\includegraphics{bar}\n"
            "\n"
            "\\end{document}\n"
        )
        self.assertEqual(self.renderer.render(Document(raw)), target)

    def test_footnote_link(self):
        from mistletoe import Document

        raw = ["[name][key]\n", "\n", "[key]: target\n"]
        target = (
            "\\documentclass{article}\n"
            "\\usepackage{hyperref}\n"
            "\\begin{document}\n"
            "\n"
            "\\href{target}{name}"
            "\n"
            "\\end{document}\n"
        )
        self.assertEqual(self.renderer.render(Document(raw)), target)



from unittest import TestCase
from unittest.mock import call, patch, sentinel, mock_open, Mock
from mistletoe import cli


class TestCLI(TestCase):
    @patch(
        "mistletoe.cli.parse",
        return_value=Mock(filenames=[], renderer=sentinel.Renderer),
    )
    @patch("mistletoe.cli.interactive")
    def test_main_to_interactive(self, mock_interactive, mock_parse):
        cli.main(None)
        mock_interactive.assert_called_with(sentinel.Renderer)

    @patch(
        "mistletoe.cli.parse",
        return_value=Mock(filenames=["foo.md"], renderer=sentinel.Renderer),
    )
    @patch("mistletoe.cli.convert")
    def test_main_to_convert(self, mock_convert, mock_parse):
        cli.main(None)
        mock_convert.assert_called_with(["foo.md"], sentinel.Renderer)

    @patch("importlib.import_module", return_value=Mock(Renderer=sentinel.RendererCls))
    def test_parse_renderer(self, mock_import_module):
        namespace = cli.parse(["-r", "foo.Renderer"])
        mock_import_module.assert_called_with("foo")
        self.assertEqual(namespace.renderer, sentinel.RendererCls)

    def test_parse_filenames(self):
        filenames = ["foo.md", "bar.md"]
        namespace = cli.parse(filenames)
        self.assertEqual(namespace.filenames, filenames)

    @patch("mistletoe.cli.convert_file")
    def test_convert(self, mock_convert_file):
        filenames = ["foo", "bar"]
        cli.convert(filenames, sentinel.RendererCls)
        calls = [call(filename, sentinel.RendererCls) for filename in filenames]
        mock_convert_file.assert_has_calls(calls)

    @patch("mistletoe.markdown", return_value="rendered text")
    @patch("builtins.print")
    @patch("builtins.open", new_callable=mock_open)
    def test_convert_file_success(self, mock_open_, mock_print, mock_markdown):
        filename = "foo"
        cli.convert_file(filename, sentinel.RendererCls)
        mock_open_.assert_called_with(filename, "r")
        mock_print.assert_called_with("rendered text", end="")

    @patch("builtins.open", side_effect=OSError)
    @patch("sys.exit")
    def test_convert_file_fail(self, mock_exit, mock_open_):
        filename = "foo"
        cli.convert_file(filename, sentinel.RendererCls)
        mock_open_.assert_called_with(filename, "r")
        mock_exit.assert_called_with('Cannot open file "foo".')

    @patch("mistletoe.cli._import_readline")
    @patch("mistletoe.cli._print_heading")
    @patch("mistletoe.markdown", return_value="rendered text")
    @patch("builtins.print")
    def test_interactive(
        self, mock_print, mock_markdown, mock_print_heading, mock_import_readline
    ):
        def MockInputFactory(return_values):
            _counter = -1

            def mock_input(prompt=""):
                nonlocal _counter
                _counter += 1
                if _counter < len(return_values):
                    return return_values[_counter]
                elif _counter == len(return_values):
                    raise EOFError
                else:
                    raise KeyboardInterrupt

            return mock_input

        return_values = ["foo", "bar", "baz"]
        with patch("builtins.input", MockInputFactory(return_values)):
            cli.interactive(sentinel.RendererCls)

        mock_import_readline.assert_called_with()
        mock_print_heading.assert_called_with(sentinel.RendererCls)
        mock_markdown.assert_called_with(
            ["foo\n", "bar\n", "baz\n"], sentinel.RendererCls
        )
        calls = [call("\nrendered text", end=""), call("\nExiting.")]
        mock_print.assert_has_calls(calls)

    @patch("importlib.import_module", return_value=Mock(Renderer=sentinel.RendererCls))
    def test_import_success(self, mock_import_module):
        self.assertEqual(sentinel.RendererCls, cli._import("foo.Renderer"))

    @patch("sys.exit")
    def test_import_incomplete_path(self, mock_exit):
        cli._import("foo")
        error_msg = "[error] please supply full path to your custom renderer."
        mock_exit.assert_called_with(error_msg)

    @patch("importlib.import_module", side_effect=ImportError)
    @patch("sys.exit")
    def test_import_module_error(self, mock_exit, mock_import_module):
        cli._import("foo.Renderer")
        mock_exit.assert_called_with('[error] cannot import module "foo".')

    @patch("importlib.import_module", return_value=Mock(spec=[]))
    @patch("sys.exit")
    def test_import_class_error(self, mock_exit, mock_import_module):
        cli._import("foo.Renderer")
        error_msg = '[error] cannot find renderer "Renderer" from module "foo".'
        mock_exit.assert_called_with(error_msg)

    @patch("builtins.__import__")
    @patch("builtins.print")
    def test_import_readline_success(self, mock_print, mock_import):
        cli._import_readline()
        mock_print.assert_not_called()

    @patch("builtins.__import__", side_effect=ImportError)
    @patch("builtins.print")
    def test_import_readline_fail(self, mock_print, mock_import):
        cli._import_readline()
        mock_print.assert_called_with("[warning] readline library not available.")

    @patch("builtins.print")
    def test_print_heading(self, mock_print):
        cli._print_heading(Mock(__name__="Renderer"))
        version = cli.mistletoe.__version__
        msgs = [
            "mistletoe [version {}] (interactive)".format(version),
            "Type Ctrl-D to complete input, or Ctrl-C to exit.",
            "Using renderer: Renderer",
        ]
        calls = [call(msg) for msg in msgs]
        mock_print.assert_has_calls(calls)



"""
Built-in span-level token classes.
"""

import re
from threading import local

import mistletoe.span_tokenizer as tokenizer
from mistletoe import core_tokens


"""
Tokens to be included in the parsing process, in the order specified.
"""
__all__ = [
    "EscapeSequence",
    "Strikethrough",
    "AutoLink",
    "CoreTokens",
    "InlineCode",
    "LineBreak",
    "RawText",
]


_root_node = None


def tokenize_inner(content):
    """
    A wrapper around span_tokenizer.tokenize. Pass in all span-level token
    constructors as arguments to span_tokenizer.tokenize.

    Doing so (instead of importing span_token module in span_tokenizer)
    avoids cyclic dependency issues, and allows for future injections of
    custom token classes.

    _token_types variable is at the bottom of this module.

    See also: span_tokenizer.tokenize, block_token.tokenize.
    """
    return tokenizer.tokenize(content, _token_types.value)


def add_token(token_cls, position=1):
    """
    Allows external manipulation of the parsing process.
    This function is called in BaseRenderer.__enter__.

    Arguments:
        token_cls (SpanToken): token to be included in the parsing process.
    """
    _token_types.value.insert(position, token_cls)


def remove_token(token_cls):
    """
    Allows external manipulation of the parsing process.
    This function is called in BaseRenderer.__exit__.

    Arguments:
        token_cls (SpanToken): token to be removed from the parsing process.
    """
    _token_types.value.remove(token_cls)


def reset_tokens():
    """
    Resets global _token_types to all token classes in __all__.
    """
    global _token_types
    _token_types.value = [globals()[cls_name] for cls_name in __all__]


class SpanToken:
    parse_inner = True
    parse_group = 1
    precedence = 5

    def __init__(self, match):
        if not self.parse_inner:
            self.content = match.group(self.parse_group)

    def __contains__(self, text):
        if hasattr(self, "children"):
            return any(text in child for child in self.children)
        return text in self.content

    @classmethod
    def find(cls, string):
        return cls.pattern.finditer(string)


class CoreTokens(SpanToken):
    precedence = 3

    def __new__(self, match):
        return globals()[match.type](match)

    @classmethod
    def find(cls, string):
        return core_tokens.find_core_tokens(string, _root_node)


class Strong(SpanToken):
    """
    Strong tokens. ("**some text**")
    """


class Emphasis(SpanToken):
    """
    Emphasis tokens. ("*some text*")
    """


class InlineCode(SpanToken):
    """
    Inline code tokens. ("`some code`")
    """

    pattern = re.compile(r"(?<!\\|`)(?:\\\\)*(`+)(?!`)(.+?)(?<!`)\1(?!`)", re.DOTALL)
    parse_inner = False
    parse_group = 2

    def __init__(self, match):
        content = match.group(self.parse_group)
        self.children = (RawText(" ".join(re.split("[ \n]+", content.strip()))),)

    @classmethod
    def find(cls, string):
        matches = core_tokens._code_matches
        core_tokens._code_matches = []
        return matches


class Strikethrough(SpanToken):
    """
    Strikethrough tokens. ("~~some text~~")
    """

    pattern = re.compile(r"(?<!\\)(?:\\\\)*~~(.+)~~", re.DOTALL)


class Image(SpanToken):
    """
    Image tokens. ("![alt](src "title")")

    Attributes:
        src (str): image source.
        title (str): image title (default to empty).
    """

    def __init__(self, match):
        self.src = match.group(2).strip()
        self.title = match.group(3)


class Link(SpanToken):
    """
    Link tokens. ("[name](target)")

    Attributes:
        target (str): link target.
    """

    def __init__(self, match):
        self.target = EscapeSequence.strip(match.group(2).strip())
        self.title = EscapeSequence.strip(match.group(3))


class AutoLink(SpanToken):
    """
    Autolink tokens. ("<http://www.google.com>")

    Attributes:
        children (iterator): a single RawText node for alternative text.
        target (str): link target.
    """

    pattern = re.compile(
        r"(?<!\\)(?:\\\\)*<([A-Za-z][A-Za-z0-9+.-]{1,31}:[^ <>]*?|[A-Za-z0-9.!#$%&'*+/=?^_`{|}~-]+@[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?(?:\.[A-Za-z0-9](?:[A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*)>"  # noqa: E501
    )
    parse_inner = False

    def __init__(self, match):
        content = match.group(self.parse_group)
        self.children = (RawText(content),)
        self.target = content
        self.mailto = "@" in self.target and "mailto" not in self.target.casefold()


class EscapeSequence(SpanToken):
    """
    Escape sequences. ("\\*")

    Attributes:
        children (iterator): a single RawText node for alternative text.
    """

    pattern = re.compile(r"\\([!\"#$%&'()*+,-./:;<=>?@\[\\\]^_`{|}~])")
    parse_inner = False
    precedence = 2

    def __init__(self, match):
        self.children = (RawText(match.group(self.parse_group)),)

    @classmethod
    def strip(cls, string):
        return cls.pattern.sub(r"\1", string)


class LineBreak(SpanToken):
    """
    Hard or soft line breaks.
    """

    pattern = re.compile(r"( *|\\)\n")
    parse_inner = False
    parse_group = 0

    def __init__(self, match):
        content = match.group(1)
        self.soft = not content.startswith(("  ", "\\"))
        self.content = ""


class RawText(SpanToken):
    """
    Raw text. A leaf node.

    RawText is the only token that accepts a string for its constructor,
    instead of a match object. Also, all recursions should bottom out here.
    """

    def __init__(self, content):
        self.content = content


_tags = {
    "address",
    "article",
    "aside",
    "base",
    "basefont",
    "blockquote",
    "body",
    "caption",
    "center",
    "col",
    "colgroup",
    "dd",
    "details",
    "dialog",
    "dir",
    "div",
    "dl",
    "dt",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "frame",
    "frameset",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "head",
    "header",
    "hr",
    "html",
    "iframe",
    "legend",
    "li",
    "link",
    "main",
    "menu",
    "menuitem",
    "meta",
    "nav",
    "noframes",
    "ol",
    "optgroup",
    "option",
    "p",
    "param",
    "section",
    "source",
    "summary",
    "table",
    "tbody",
    "td",
    "tfoot",
    "th",
    "thead",
    "title",
    "tr",
    "track",
    "ul",
}

_tag = r"[A-Za-z][A-Za-z0-9-]*"
_attrs = r'(?:\s+[A-Za-z_:][A-Za-z0-9_.:-]*(?:\s*=\s*(?:[^ "\'=<>`]+|\'[^\']*?\'|"[^\"]*?"))?)*'  # noqa: E501

_open_tag = r"(?<!\\)<" + _tag + _attrs + r"\s*/?>"
_closing_tag = r"(?<!\\)</" + _tag + r"\s*>"
_comment = r"(?<!\\)<!--(?!>|->)(?:(?!--).)+?(?<!-)-->"
_instruction = r"(?<!\\)<\?.+?\?>"
_declaration = r"(?<!\\)<![A-Z].+?>"
_cdata = r"(?<!\\)<!\[CDATA.+?\]\]>"


class HTMLSpan(SpanToken):
    """
    Span-level HTML tokens.

    Attributes:
        content (str): literal strings rendered as-is.
    """

    pattern = re.compile(
        "|".join(
            [_open_tag, _closing_tag, _comment, _instruction, _declaration, _cdata]
        ),
        re.DOTALL,
    )
    parse_inner = False
    parse_group = 0


_token_types = local()
_token_types.value = []
reset_tokens()



import re
import mistletoe.span_token as span_token


__all__ = ["Math"]


class Math(span_token.SpanToken):
    pattern = re.compile(r"(\${1,2})([^$]+?)\1")
    parse_inner = False
    parse_group = 0



from setuptools import setup
import mistletoe

setup(
    name="mistletoe-ebp",
    version=mistletoe.__version__,
    description="A fast, extensible Markdown parser in pure Python.",
    url="https://github.com/miyuchina/mistletoe",
    author="Mi Yu",
    author_email="hello@afteryu.me",
    license="MIT",
    packages=["mistletoe"],
    entry_points={"console_scripts": ["mistletoe = mistletoe.__main__:main"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Markup",
    ],
    keywords="markdown lexer parser development",
    python_requires="~=3.5",
    extras_require={
        "code_style": ["flake8<3.8.0,>=3.7.0", "black==19.10b0", "pre-commit==1.17.0"],
        "testing": ["coverage", "pytest>=3.6,<4", "pytest-cov"],
    },
    zip_safe=False,
)
